#ifndef CPPRUBY_H
#define CPPRUBY_H

#include "CppRubyExceptionHandler.h"
#include "RubyCppTools.h"


namespace CppRuby
{


class Klass;
class String;


class Object
{
public:
  Object( VALUE instance )
    : m_instance( instance )
  {
  }

  VALUE instance()
  {
    return m_instance;
  }

  operator VALUE()
  {
    return instance();
  }

  bool isNil()
  {
    return instance() == Qnil;
  }

  Klass type();

  static ID methodId( const char *name )
  {
    return rb_intern( name );
  }

  bool asBool()
  {
    return !( m_instance == Qnil  ||  m_instance == Qfalse );
  }

  String inspect();

  Object methodCall( ID methodId )
  {
    return methodCall( methodId, 0, NULL );
  }

  Object methodCall( const char *name )
  {
    return methodCall( methodId(name), 0, NULL );
  }

  Object methodCall( ID methodId, VALUE argument1 )
  {
    VALUE arguments[] = { argument1 };
    return methodCall( methodId, 1, arguments );
  }

  Object methodCall( const char *name, VALUE argument1 )
  {
    return methodCall( methodId(name), argument1 );
  }

  Object methodCall( ID methodId, VALUE argument1, VALUE argument2 )
  {
    VALUE arguments[] = { argument1, argument2 };
    return methodCall( methodId, 2, arguments );
  }

  Object methodCall( const char *name, VALUE argument1, VALUE argument2 )
  {
    return methodCall( methodId(name), argument1, argument2 );
  }

  Object methodCall( ID methodId, VALUE argument1, VALUE argument2, VALUE argument3 )
  {
    VALUE arguments[] = { argument1, argument2, argument3 };
    return methodCall( methodId, 3, arguments );
  }

  Object methodCall( const char *name, VALUE argument1, VALUE argument2, VALUE argument3 )
  {
    return methodCall( methodId(name), argument1, argument2, argument3 );
  }

  Object methodCall( ID methodId,
                     int argumentCount,
                     VALUE arguments[] )
  {
    // should check if m_instance is Qnil ?
    return rb_funcall2( m_instance, methodId, argumentCount, arguments );
  }

  Object safeMethodCall( ID methodId,
                         ExceptionHandler &handler = DefaultExceptionHandler() )
  {
    return safeMethodCall( methodId, 0, NULL, handler );
  }

  Object safeMethodCall( const char *name,
                         ExceptionHandler &handler = DefaultExceptionHandler() )
  {
    return safeMethodCall( methodId(name), handler );
  }

  Object safeMethodCall( ID methodId,
                         VALUE argument1,
                         ExceptionHandler &handler = DefaultExceptionHandler() )
  {
    VALUE arguments[] = { argument1 };
    return safeMethodCall( methodId, 1, arguments, handler );
  }

  Object safeMethodCall( const char *name,
                         VALUE argument1,
                         ExceptionHandler &handler = DefaultExceptionHandler() )
  {
    return safeMethodCall( methodId(name), argument1, handler );
  }

  Object safeMethodCall( ID methodId,
                         int argumentCount,
                         VALUE arguments[],
                         ExceptionHandler &handler = DefaultExceptionHandler() )
  {
    return protect2( SafeMethodCallFunctor( *this, 
                                            methodId), 
                     argumentCount, 
                     arguments,
                     handler );
  }

  String to_s();

protected:
  struct SafeMethodCallFunctor
  {
    SafeMethodCallFunctor( Object &object, ID methodId )
      : m_object( object)
      , m_methodId( methodId )
    {
    }

    Object operator()( int argumentCount,
                       VALUE arguments[] )
    {
      return m_object.methodCall( m_methodId, argumentCount, arguments );
    }

    Object &m_object;
    ID m_methodId;
  };

  VALUE m_instance;
};


/* Creating a new instance:
VALUE rb_cPoint = rb_const_get(rb_cObject, rb_intern("Point"));
  return rb_funcall(rb_cPoint, rb_intern("new"), INT2NUM(3), INT2NUM(4));

or :

  VALUE args[] = { INT2NUM(3), INT2NUM(4) };
  return rb_class_new_instance(2, args, rb_cPoint);
*/



inline VALUE 
getClass( const char *className )
{
  return rb_const_get( rb_cObject, rb_intern( className ) );
}


class Klass : public Object
{
public:
  Klass( const char *name )
    : Object( getClass( name ) )
  {
  }

  Klass( VALUE klass )
    : Object( klass )
  {
  }

  VALUE klass()
  {
    return m_instance;
  }

  Object newInstance()
  {
    return newInstance( 0, NULL );
  }

  Object newInstance( VALUE argument1 )
  {
    VALUE arguments[1] = { argument1 };
    return newInstance( 1, arguments );
  }

  Object newInstance( VALUE argument1, VALUE argument2 )
  {
    VALUE arguments[2] = { argument1, argument2 };
    return newInstance( 2, arguments );
  }

  Object newInstance( VALUE argument1, VALUE argument2, VALUE argument3 )
  {
    VALUE arguments[3] = { argument1, argument2, argument3 };
    return newInstance( 3, arguments );
  }

  Object newInstance( int argCount,
                      VALUE arguments[] )
  {
    return rb_class_new_instance( argCount, arguments, m_instance );
  }
};


inline Klass 
Object::type()
{
  return CLASS_OF( m_instance );
}


class String : public Object
{
public:
  String( const char *str )
    : Object( rb_str_new2( str ) )
  {
  }

  String( VALUE str )
    : Object( str )
  {
  }

  const char *c_str()
  {
    return m_instance != Qnil ? STR2CSTR( m_instance ) : "";
  }

  VALUE str()
  {
    return m_instance;
  }

  operator const char *()
  {
    return c_str();
  }
};


inline String 
Object::to_s()
{
  return methodCall( "to_s" );
}


inline String 
Object::inspect()
{
  return rb_inspect( m_instance );
}


class Array : public Object
{
public:
  Array( VALUE array )
    : Object( array )
  {
  }

  static Array newInstance()
  {
    return rb_ary_new();
  }

  Array push( VALUE value )
  {
    return methodCall( "push", value );
  }

  String join( const char *str )
  {
    return methodCall( "join", String(str) );
  }

  String join( VALUE str )
  {
    return methodCall( "join", str );
  }
};


/* Getting my method name (use this in a ruby extension method implementation.
    char *name = rb_id2name(rb_frame_last_func());
    return rb_str_new2(name);
 */



/* Default parameter in method declaration:
Baptiste Lepilleur wrote:

>   rb_define_method( ruby_cScintillaBase,
>                       "sendEditorMessage",
>                       ruby_sendEditorMessage,
>                       3 );
> 
>     How do I define default value for parameter 2 & 3 ? Is it even possible
> to do that from C ?

You can use rb_scan_args. See the "Extending Ruby" chapter of the
pickaxe book, and the README.EXT that came with your Ruby. Note that you
have to use an arg count of -1, so that the args will be passed in a C
array, which you pass to rb_scan_args, along with a format string.

In the CGenerator library in RAA, this is made even easier. When you
define a C function from Ruby, you specify something like

  c_array_args {
    required :arg0, :arg1
    optional :arg2, :arg3, :arg4
    rest     :rest
    block    :block
  }

This causes code to be generated using rb_scan_args that parses the
arguments into the C variables that you named with the symbols. You can
then access them in the body of your C function.

You can specify defaults by adding the following to the block above:

    default   :arg3 => "INT2NUM(7)",
              :arg4 => "INT2NUM(NUM2INT(arg2) + NUM2INT(arg3))"

The stuff between the quotes is C code.

You can even specify typechecking of your arguments:

    typecheck :arg2 => Numeric, :arg3 => Numeric

See examples/sample.rb for a simple, but complete example.
*/



} // namespace CppRuby

#endif