#define WIN32_LEAN_AND_MEAN 
#define NOGDI
#define NOKERNEL
#define NOSOUND
#define NOMINMAX
#include <windows.h>

#include <ruby.h>
#include "RubyExtScintilla.h"
#include "RubyTools.h"


extern "C" {

void Init_ScintillaNotification();


VALUE ruby_cScintillaBase;
VALUE ruby_cScintillaBridge;
ID ruby_editorWindowHandle;


void *
Get_Scintilla_Window_Handle( VALUE editor )
{
 return (HWND)NUM2INT( rb_ivar_get( editor, ruby_editorWindowHandle ) );
}


static VALUE
ruby_sendStyledTextEditorMessage( int from,
                                  int to,
                                  HWND editorWindowHandle )
{
  if ( to <= from )
    return rb_ary_new3( 2, rb_str_new( "", 0 ), rb_ary_new() );

  TextRange range;
  range.chrg.cpMin = from;
  range.chrg.cpMax = to;
  range.lpstrText = ALLOC_N( char, (to - from +2)*2 );
  int actualLength = SendMessage( editorWindowHandle, 
                                  SCI_GETSTYLEDTEXT, 
                                  0,
                                  (LPARAM)&range ) / 2;

  VALUE styles = rb_ary_new2( actualLength );
  VALUE text = rb_str_new( range.lpstrText, actualLength );

  const char *index = range.lpstrText;
  const char *endIndex = index + actualLength*2;
  char *destText = RSTRING(text)->ptr;
  while ( index < endIndex )
  {
    *destText++ = *index++;
    rb_ary_push( styles, INT2FIX( (unsigned int)*index++) );
  }

  xfree( range.lpstrText );
  return rb_ary_new3( 2, text, styles );
}


static VALUE
ruby_sendTextRangeEditorMessage( int from,
                                 int to,
                                 HWND editorWindowHandle )
{
  if ( to <= from )
    return rb_str_new( "", 0 );

  TextRange range;
  range.chrg.cpMin = from;
  range.chrg.cpMax = to;
  range.lpstrText = ALLOC_N( char, to - from +2 );
  int actualLength = SendMessage( editorWindowHandle, 
                                  SCI_GETTEXTRANGE, 
                                  0,
                                  (LPARAM)&range );
  VALUE text = rb_str_new( range.lpstrText, actualLength );
  xfree( range.lpstrText );
  return text;
}


static VALUE 
ruby_sendStringResultEditorMessage( int textLength,
                                    UINT message,
                                    HWND editorWindowHandle,
                                    int lineNumber = -1 )
{
  if ( textLength <= 0 )
    return rb_str_new( "", 0 );

  char *buffer = ALLOC_N( char, textLength +2 );
  int actualLength = SendMessage( editorWindowHandle, 
                                  message, 
                                  (WPARAM)(lineNumber != -1 ? lineNumber
                                                            : (textLength +1) ),
                                  (LPARAM)buffer );
  VALUE text = rb_str_new( buffer, actualLength );
  xfree( buffer );
  return text;
}


static VALUE 
ruby_sendEditorMessage( VALUE self,
                        VALUE code,
                        VALUE param1,
                        VALUE param2 )
{
  Check_Type( code, T_FIXNUM );
  UINT message = FIX2UINT(code);

  HWND editorWindowHandle = (HWND)Get_Scintilla_Window_Handle( self );

  WPARAM wParam;
  if ( FIXNUM_P( param1 ) )
    wParam = FIX2UINT( param1 );
  else if ( param1 == Qfalse  ||  param1 == Qnil )
    wParam = 0;
  else if ( param1 == Qtrue )
    wParam = 1;
  else 
    wParam = NUM2UINT( param1 );


  switch ( message )    // Handle Messages that return a string (stringresult)
  {
  case SCI_GETTEXTRANGE:
    {
      return ruby_sendTextRangeEditorMessage( wParam,
                                              NUM2INT(param2), 
                                              editorWindowHandle );
    }
  case SCI_GETSTYLEDTEXT:
    {
      return ruby_sendStyledTextEditorMessage( wParam,
                                               NUM2INT(param2), 
                                               editorWindowHandle );
    }
  case SCI_GETCURLINE:
    {  // get the line number at the current position, then get its length
      int currentPosition = SendMessage( editorWindowHandle, 
                                         SCI_GETCURRENTPOS, 0, 0 );
      int currentLineNumber = SendMessage( editorWindowHandle,  
                                           SCI_LINEFROMPOSITION, 
                                           currentPosition, 0);
      int lineLength = SendMessage( editorWindowHandle, 
                                    SCI_LINELENGTH, 
                                    currentLineNumber, 0 );
      return ruby_sendStringResultEditorMessage( lineLength, 
                                                 SCI_GETCURLINE, 
                                                 editorWindowHandle );
    }
  case SCI_GETLINE:
    { // get the line length and pass it, as well as the line number (special case)
      int lineLength = SendMessage( editorWindowHandle, 
                                    SCI_LINELENGTH, 
                                    wParam, 0 );
      return ruby_sendStringResultEditorMessage( lineLength, 
                                                 SCI_GETLINE, 
                                                 editorWindowHandle,
                                                 wParam );
    }
  case SCI_GETTEXT:
    {
      int textLength = SendMessage( editorWindowHandle, SCI_GETTEXTLENGTH, 0, 0 );
      return ruby_sendStringResultEditorMessage( textLength, 
                                                 SCI_GETTEXT,
                                                 editorWindowHandle );
    }
  case SCI_GETSELTEXT:
    {
      int selStartPos = SendMessage( editorWindowHandle, SCI_GETSELECTIONSTART, 0, 0 );
      int selEndPos = SendMessage( editorWindowHandle, SCI_GETSELECTIONEND, 0, 0 );
      return ruby_sendStringResultEditorMessage( selEndPos - selStartPos, 
                                                 SCI_GETSELTEXT,
                                                 editorWindowHandle );
    }
  }

  LPARAM lParam;
  switch ( TYPE( param2 ) )
  {
  case T_FIXNUM:
    lParam = FIX2UINT( param2 );
    break;
  case T_FALSE:
  case T_NIL:
    lParam = 0;
    break;
  case T_TRUE:
    lParam = 1;
    break;
  case T_BIGNUM:
    lParam = NUM2UINT( param2 );
    break;
  case T_STRING:
    lParam = (LPARAM)STR2CSTR( param2 );
    break;
  default:    // force an exception to be raised...
    rb_check_type( param2, T_FIXNUM );
  }

  LRESULT result = SendMessage( editorWindowHandle, 
                                message,
                                wParam, 
                                lParam );
  return INT2NUM( result );
}


static void
ruby_updateClassScintillaBase()
{
  ruby_cScintillaBase = rb_define_class( "ScintillaBase", rb_cObject );
  ruby_editorWindowHandle = rb_intern( "@editorWindowHandle" );
  ruby_define_method( ruby_cScintillaBase, 
                      "sendEditorMessage", 
                      ruby_sendEditorMessage,
                      3 );
}


void
Init_Scintilla()
{
  ruby_load( "ScintillaBase.rb" );
  ruby_load( "ScintillaBridge.rb" );

  ruby_updateClassScintillaBase();

  ruby_cScintillaBridge = ruby_get_class( "ScintillaBridge" );

  Init_ScintillaNotification();
}



static VALUE
Scintilla_Notification_GetField_eventType( VALUE self )
{
  SCNotification *message;
  Data_Get_Struct( self, SCNotification, message );
  return INT2NUM( message->nmhdr.code );
}


static VALUE
Scintilla_Notification_GetField_ch( VALUE self )
{
  SCNotification *message;
  Data_Get_Struct( self, SCNotification, message );
  char c = message->ch;
  return rb_str_new( &c, 1 );
}


static VALUE
Scintilla_Notification_GetField_text( VALUE self )
{
  SCNotification *message;
  Data_Get_Struct( self, SCNotification, message );
  return message->text ? rb_str_new2( message->text ) : rb_str_new( "", 0 );
}


#undef SCINTILLA_NOTIFICATION_GETFIELD_INT
#define SCINTILLA_NOTIFICATION_GETFIELD_INT( fieldName )                  \
  static VALUE Scintilla_Notification_GetField_##fieldName( VALUE self )  \
  {                                                                       \
    SCNotification *message;                                              \
    Data_Get_Struct( self, SCNotification, message );                     \
    return INT2NUM( message->fieldName );                                 \
  }

SCINTILLA_NOTIFICATION_GETFIELD_INT( position )
SCINTILLA_NOTIFICATION_GETFIELD_INT( modifiers )
SCINTILLA_NOTIFICATION_GETFIELD_INT( modificationType )
SCINTILLA_NOTIFICATION_GETFIELD_INT( length )
SCINTILLA_NOTIFICATION_GETFIELD_INT( linesAdded )
SCINTILLA_NOTIFICATION_GETFIELD_INT( message )
SCINTILLA_NOTIFICATION_GETFIELD_INT( wParam )
SCINTILLA_NOTIFICATION_GETFIELD_INT( lParam )
SCINTILLA_NOTIFICATION_GETFIELD_INT( line )
SCINTILLA_NOTIFICATION_GETFIELD_INT( foldLevelNow )
SCINTILLA_NOTIFICATION_GETFIELD_INT( foldLevelPrev )
SCINTILLA_NOTIFICATION_GETFIELD_INT( margin )
SCINTILLA_NOTIFICATION_GETFIELD_INT( listType )
SCINTILLA_NOTIFICATION_GETFIELD_INT( x )
SCINTILLA_NOTIFICATION_GETFIELD_INT( y )


static void
Free_Scintilla_Notification( void *notification )
{
  xfree( notification );    // is that correct. It is not documented ???
}


void 
Scintilla_Forward_Notification( SCNotification *notification, 
                                VALUE target )
{
  if ( target == Qnil )
    return;

  VALUE classScintillaNotification = ruby_get_class( "ScintillaNotification" );

  SCNotification *message = ALLOC( SCNotification );
  memcpy ( message, notification, sizeof(SCNotification) );
  switch( notification->nmhdr.code )    // handle special field duplication (string)
  {
  case SCN_MODIFIED:
    message->text = strdup( notification->text );
    break;
  // notes: SCN_MACRORECORD have wParam & lParam fields which may need to be
  // duplicated depending on the nature of the field. Investigate later.
  // Don't support this event for now.
  case SCN_CHARADDED:
    break;
  default:
    message->text = NULL;
    break;
  }

  VALUE messageObject = Data_Wrap_Struct( classScintillaNotification,
                                          0,
                                          Free_Scintilla_Notification,
                                          message );
  rb_funcall( target, rb_intern( "scintillaEvent" ), 1, messageObject );
}


void 
Init_ScintillaNotification()
{
  VALUE classScintillaNotification = rb_define_class( "ScintillaNotification", 
                                                      rb_cObject );

#undef SCINTILLA_NOTIFICATION_DEFINE_FIELD_METHOD
#define SCINTILLA_NOTIFICATION_DEFINE_FIELD_METHOD( fieldName )     \
  ruby_define_method( classScintillaNotification,                   \
                      #fieldName,                                   \
                      Scintilla_Notification_GetField_##fieldName,  \
                      0 )
  
  SCINTILLA_NOTIFICATION_DEFINE_FIELD_METHOD( eventType );
  SCINTILLA_NOTIFICATION_DEFINE_FIELD_METHOD( ch );
  SCINTILLA_NOTIFICATION_DEFINE_FIELD_METHOD( position );
  SCINTILLA_NOTIFICATION_DEFINE_FIELD_METHOD( modifiers );
  SCINTILLA_NOTIFICATION_DEFINE_FIELD_METHOD( modificationType );
  SCINTILLA_NOTIFICATION_DEFINE_FIELD_METHOD( text );
  SCINTILLA_NOTIFICATION_DEFINE_FIELD_METHOD( length );
  SCINTILLA_NOTIFICATION_DEFINE_FIELD_METHOD( linesAdded );
  SCINTILLA_NOTIFICATION_DEFINE_FIELD_METHOD( message );
  SCINTILLA_NOTIFICATION_DEFINE_FIELD_METHOD( wParam );
  SCINTILLA_NOTIFICATION_DEFINE_FIELD_METHOD( lParam );
  SCINTILLA_NOTIFICATION_DEFINE_FIELD_METHOD( line );
  SCINTILLA_NOTIFICATION_DEFINE_FIELD_METHOD( foldLevelNow );
  SCINTILLA_NOTIFICATION_DEFINE_FIELD_METHOD( foldLevelPrev );
  SCINTILLA_NOTIFICATION_DEFINE_FIELD_METHOD( margin );
  SCINTILLA_NOTIFICATION_DEFINE_FIELD_METHOD( listType );
  SCINTILLA_NOTIFICATION_DEFINE_FIELD_METHOD( x );
  SCINTILLA_NOTIFICATION_DEFINE_FIELD_METHOD( y );
}


}   // extern "C"
