# Scintilla Bridge Generator
# Generate ruby wrapper for Scintilla
# - parsing:
#   main driver is parseIFace()
#   parsing of function (get/set/fun) is in Function#initialize, throw
#       UnsupportedFunctionError if the function is not supported (no
#       mapping for this type)
#   parsing of event is in Event#initialize
# - code generation
#   main driver is generate()
#     delegate generation to Function#generate and Event#generate
#
#
# Rules to generate the code
# - unused parameters disappears
# - type mapping is done with function defined in Function#setUpTypeMapper.
#   The asXXX() function are defined in ScintillaBase.
#   int => Numeric
#   string => any object with to_s().
#   bool => any object. nil and false are false, anything else is true.
#   position => equivalent to int
#   colour => ScintillaBase#Color
#   stringresult => method with that type have been modified to return the
#                   string as result instead of passing a preallocated buffer
#                   as parameter. int GetText(int length, stringresult text)
#                   is now string GetText()... length parameter is removed
#                   from the signature (not needed)
#   textrange => use specific method defined in ScintillaBase, usually 
#                arguments are spliited, and the string is returned as result.
#   keymod => use specific method defined in ScintillaBase.
# - getter function name: the 'get' is removed: GetTextLength() is textLength().
# - setter function name: the 'set' is removed and '=' is appended if
#   the function as only one parameter: SetOvertype(bool) is overtype=(bool), but
#   StyleSetFore(int,color) is styleSetFore(int,color).
# - if a function or a getter retursn bool, then '?' is append to the name:
#   CanRedo() is canRedo?(), GetUseTabs() is useTabs?()
# - if the function is already defined in ScintillaBase, then Message is appended
#   to the function name. ScintillaBase defines addText( string ), the generated
#   function will be addTextMessage( int, string). [addText() forward to
#   addTextMessage() after extracting the text length].
#
# To Do: 
# - Checks that the stringresult function is a supported function (code is handled
# on the C side)
# - Removes warning for function implemented in ScintillaBase.
#
#

require 'ScintillaBase.rb'


class ScintillaBridgeGenerator
  INDEX_FEATURE_TYPE = 0
  INDEX_CONSTANT_DECL = 1
  INDEX_RETURN_TYPE = 2
  INDEX_FUNCTION_DECL = 3
  
  class UnsupportedFunctionError < Exception
    attr_reader :unsupportedType
    
    def initialize( unsupportedType )
      super( "Unsupported function" )
      @unsupportedType = unsupportedType
    end
  end

  class Event
    def initialize( declaration )
      # evt void <name>=<number(<fieldName>,<fieldName>,...)
      match = /\s?evt\s+void\s+(\w+)=(\d+)\((.*)\)\s?/.match( declaration )
      @name = match[1]
      @code = match[2]
      @fields = match[3].split(',')
      @fields.delete 'void'
      @fields.collect! do |field|
        field.split( ' ' )[1]
      end
    end

    def generateDispatch( file )
      file.puts '      when %d' % @code
      file.print'        event' + @name + '('
      separator = ' '
      for field in @fields
        file.print separator + 'notification.' + field
        separator = ', '
      end
      file.print ' ' if separator != ' '
      file.puts ')'
    end

    def generateAdapter( file )
      file.print '  def event' + @name
      separator = '( '
      for field in @fields
        file.print separator + field
        separator = ', '
      end
      file.print ' )' if @fields.length > 0
      file.puts
      file.puts '  end'
    end
  end

  class Function
  
    def initialize( declaration, comment )
      setUpTypeMapper()
      # <featureType><ws><returnType><ws><name>[=<number](<param>,<param>)
      @comment = comment
      match = /\s?\w+\s+(\w+)\s+(\w+)=(\d+)\((.*)\)\s?/.match( declaration )
      @returnType = match[1]
      @name = match[2]
      @code = match[3]
=begin      
      puts "Function declaration:"  <<  declaration
      puts "  Function name:"  <<  @name
      puts "  Function returnType:"  <<  @returnType
      puts "  Function code:"  <<  @code
      print "  Function parameters:"  
      puts  match[4].to_s
=end
      checkSupportedType( @returnType )

      @parameters = Array.new
      match[4].strip().split( ',' ).each do |parameterDeclaration|
        parameter = parameterDeclaration.split(' ') #   [type, name]
        checkSupportedType( parameter[0] ) if parameter[0]
        parameter[1] = decorateReservedKeyWords( parameter[1] ) if parameter[1]
        @parameters  <<  parameter
      end

	  modifyStringResultFunctionSignature()
      
      normalizeName()
    end

	def modifyStringResultFunctionSignature
	  if @parameters.length == 2  and  @parameters[1][0] == "stringresult"
	    @parameters.pop()
	    @parameters.clear() if @parameters[0][1] == "length"
	    @returnType = 'string'
	  end
	end
    
    def normalizeName
      @name = @name.sub!( /^./, @name.downcase[0..0] )  # first letter to lower case
	  @name << '?' if @returnType == 'bool'
    end

    def setUpTypeMapper
      @typeMapper = {
        "void" => "nil",
        "int"  => "asInt( %s )",
        "bool" => "asBoolean( %s )",
        "position" => "asInt( %s )",
        "string" => "asString( %s )",
        "colour" => "asColor( %s )",
		"stringresult" => ".....Never In Source....."
      }
    end
    
    def checkSupportedType( type )
      raise UnsupportedFunctionError, type unless @typeMapper[ type ] 
    end
    
    def decorateReservedKeyWords( parameterName )
      if %w( for in if then else end def class module ).include?( parameterName )
        parameterName + "_"
      else
        parameterName
      end
    end
  
    def generateComment()
      comment = "  # "
      comment += @comment + ": " if @comment
      parameters = Array.new
      for parameter in @parameters
        parameters << parameter.join(' ') 
      end
      comment += @returnType + "(" + parameters.join(',') + ")"
      comment
    end
    
    def generate( file )
#      puts generateComment() << " " << @name
      file.puts generateComment()
      file.print "  def " << @name
	  file.print "Message" if ScintillaBase.method_defined? @name
      file.print "( " unless @parameters.empty?
      separator = ""
      for parameter in @parameters do
    next unless parameter[1]
    file.print separator << parameter[1]
    separator = ", "
      end
      file.print " )" unless @parameters.empty?
      file.puts ""
      generateMethodBody( file )
      file.puts "  end"
      file.puts ""
    end
    
    def generateMethodBody( file )
      file.print "    sendEditorMessage( #{@code}"
#no default parameter...
#      if ( definedParameterCount == 0)
#      else
        for index in 0..1
          file.print ", "
          if index > @parameters.length
            file.print "nil"
          else
            type, name = @parameters[index]
            file.print type ? @typeMapper[ type ] % name : "nil"
          end
        end
#      end
      file.puts " )"
    end
    
    def definedParameterCount
      count = 0
      for parameter in @parameters
        count += 1 if parameter[0]
      end
      count
    end
  end

  class Getter < Function
    def initialize( declaration, comment )
      super
    end
    
    def normalizeName
      @name = @name.sub( /^Get/, '' )
      @name = @name.sub!( /^./, @name.downcase[0..0] )  # first letter to lower case
	  @name << '?' if @returnType == 'bool'
    end
  end

  class Setter < Function
    def initialize( declaration, comment )
      super
    end
    
    def normalizeName
      @name = @name.sub( /^Set/, '' ) + "=" if definedParameterCount() == 1
      @name = @name.sub!( /^./, @name.downcase[0..0] )  # first letter to lower case
    end
  end
    
  def initialize( ifaceFileName )
    @constants = Array.new
    @functions = Array.new
    @events = Array.new
    parseIFace( ifaceFileName )
  end
  
  def parseIFace( ifaceFileName )
    File.open( ifaceFileName, 'r' ) do |file|
      lastComment = nil
      file.each_line do |line|
        comment = lastComment
        lastComment = nil
        fields = line.split( ' ' )
        begin
          case fields[INDEX_FEATURE_TYPE]
          when 'val'    # constant
            name,value = fields[INDEX_CONSTANT_DECL].split( '=' )
            next if name == nil or value == nil
            @constants << [name, value]
          when '#'        # comment
            lastComment = comment ? comment + " " : ""
            lastComment += line[(line.index('#')+1)..-1].strip
          when 'get'      # getter
            @functions  <<  Getter.new( line, comment )
          when 'set'      # setter
            @functions  <<  Setter.new( line, comment )
          when 'fun'      # function
            @functions  <<  Function.new( line, comment )
          when 'evt'      # event
            @events << Event.new( line )
          end
        rescue UnsupportedFunctionError
          puts "Warning, unsupported function: " + line
        end
      end
    end
  end
      
  def generate( rubyFileName )
    File.open( rubyFileName, 'w' ) do |file|
      file.puts "#This class has been generated with ScintillaBridgeGenerate.rb, do not edit!"
      file.puts ""
      file.puts "require 'ScintillaBase.rb'"
      file.puts ""
      file.puts "class ScintillaBridge < ScintillaBase"
        
      for constant in @constants
        file.puts "  #{constant[0]} = #{constant[1]}"
      end

      for function in @functions
        function.generate( file )
      end
      file.puts
      file.puts '  def scintillaEvent( notification )'
      file.puts '    case notification.eventType'

      for event in @events
        event.generateDispatch( file )
      end

      file.puts '    end'
      file.puts '  end'

      for event in @events
        file.puts
        event.generateAdapter( file )
      end
  
      file.puts 'end'
    end
  end


end


generator = ScintillaBridgeGenerator.new( "Scintilla.iface" )
generator.generate( "ScintillaBridge.rb" )
